//
//  GifMaker.h
//  GifMaker
//
//  Copyright 2006 T. Endo
//
// This file is distributed under General Public Licence version 2 (GPLv2).
//

#import <Cocoa/Cocoa.h>
#import <WebKit/WebKit.h>

// GIF image exporting class

@interface GifMaker : NSObject {
	int width;
	int height;
	int delay;
	int smoothness;
	int vibration;
	BOOL transparent;
	BOOL processing;
	int shadow[ 3 ];
	int shadowOffset[ 2 ];
	int background[ 3 ];
	int framesPerLoop;
	int maxLineDivision;
}

// initializer
-(id)init;
-(id)initWithWebView: ( WebView*) view;
// destructor
-(void)dealloc;
// interfaces to WebView
-(void)windowScriptObjectAvailable:(WebScriptObject*)ws;
+(NSString*)webScriptNameForSelector:(SEL)aSel;
+(BOOL)isSelectorExcludedFromWebScript:(SEL)aSel;
+(BOOL)isKeyExcludedFromWebScript:(const char*)k;


//-(void)resolveColor: (NSString*)text rgb:(int*)values;
// set the size of GIF image
-(void)setSize: (int)width :(int)heihgt;
// set delay time of animation. 1 means 1/100 seconds.
-(void)setDelay: (int)delayCount;
// set smoothness of lines. values more than 1 skips intermediate points.
-(void)setSmoothness: (int) smoothness;
// set vibration degree. Larger the value is the more animation wobbles.
-(void)setVibration: (int)vibration;
// set background color
-(void)setBackgroundColor: (int)r :(int)g :(int)b;
// set shadow color 
-(void)setShadowColor: (int)r: (int)g :(int)b;
// set offset distance of shadows
-(void)setShadowOffset: (int)x :(int)y;
// transparent image (not work)
-(void)setTransparent: (BOOL) flag;
// show intermediate drawings
-(void)showProcessing:(BOOL) flag;
// frames per loop
-(void)setFramesPerLoop: (int)frames;

// convert an object given by widget script to an array of GifLine objects
-(NSArray*) convertToLines: (WebScriptObject*) wso;
// this method is called by a widget and returns a temporary filename of a generated GIF image.
-(NSString*)outputWobblingGif : (WebScriptObject*) lines;
// this method provides removing way of local files, especially generated GIF images.
-(void)removeFile: (NSString*) filename;

// this method provides a temporary filename for user's temporary folder.
// Take care that the filename is not atomic
+(NSString*) createTemporaryFile : (NSString*) suffix;
@end

